@@ -8,12 +8,9 @@
 #include "pgrn-value.h"
 
 #include <access/heapam.h>
-#ifdef PGRN_SUPPORT_TABLEAM
-#	include <access/tableam.h>
-#else
-#	include <access/htup_details.h>
-#endif
+#include <access/tableam.h>
 #include <access/xact.h>
+#include <access/xlog.h>
 #include <catalog/pg_database.h>
 #include <executor/spi.h>
 #include <miscadmin.h>
@@ -41,6 +38,8 @@ PG_MODULE_MAGIC;
 
 extern PGDLLEXPORT void _PG_init(void);
 extern PGDLLEXPORT void
+pgroonga_crash_safer_reset_position_one(Datum datum) pg_attribute_noreturn();
+extern PGDLLEXPORT void
 pgroonga_crash_safer_reindex_one(Datum datum) pg_attribute_noreturn();
 extern PGDLLEXPORT void
 pgroonga_crash_safer_flush_one(Datum datum) pg_attribute_noreturn();
@@ -55,6 +54,7 @@ static volatile sig_atomic_t PGroongaCrashSaferGotSIGUSR1 = false;
 static int PGroongaCrashSaferFlushNaptime = 60;
 static char *PGroongaCrashSaferLogPath;
 static int PGroongaCrashSaferLogLevel = GRN_LOG_DEFAULT_LEVEL;
+static int PGroongaCrashSaferMaxRecoveryThreads = 0;
 PGRN_DEFINE_LOG_LEVEL_ENTRIES(PGroongaCrashSaferLogLevelEntries);
 static const char *PGroongaCrashSaferLibraryName = "pgroonga_crash_safer";
 
@@ -119,18 +119,90 @@ pgroonga_crash_safer_sigusr1(SIGNAL_ARGS)
 }
 
 static void
-pgroonga_crash_safer_reindex_one_on_exit(int code, Datum databaseInfoDatum)
+pgroonga_crash_safer_prepare_one_on_exit(int code, Datum databaseInfoDatum)
 {
 	uint64 databaseInfo = DatumGetUInt64(databaseInfoDatum);
 	Oid databaseOid;
 	Oid tableSpaceOid;
 	PGRN_DATABASE_INFO_UNPACK(databaseInfo, databaseOid, tableSpaceOid);
-	pgrn_crash_safer_statuses_set_reindex_pid(NULL,
+	pgrn_crash_safer_statuses_set_prepare_pid(NULL,
 											  databaseOid,
 											  tableSpaceOid,
 											  InvalidPid);
 }
 
+void
+pgroonga_crash_safer_reset_position_one(Datum databaseInfoDatum)
+{
+	uint64 databaseInfo = DatumGetUInt64(databaseInfoDatum);
+	Oid databaseOid;
+	Oid tableSpaceOid;
+
+	PGRN_DATABASE_INFO_UNPACK(databaseInfo, databaseOid, tableSpaceOid);
+
+	BackgroundWorkerInitializeConnectionByOid(databaseOid, InvalidOid, 0);
+
+	StartTransactionCommand();
+	SPI_connect();
+	PushActiveSnapshot(GetTransactionSnapshot());
+	pgstat_report_activity(STATE_RUNNING, TAG ": resetting position");
+
+	pgrn_crash_safer_statuses_set_prepare_pid(NULL,
+											  databaseOid,
+											  tableSpaceOid,
+											  MyProcPid);
+	before_shmem_exit(pgroonga_crash_safer_prepare_one_on_exit,
+					  databaseInfoDatum);
+
+	{
+		int result;
+
+		SetCurrentStatementStartTimestamp();
+		result = SPI_execute("SELECT proname "
+							 "  FROM pg_catalog.pg_proc "
+							 "  WHERE "
+							 "    proname = 'pgroonga_wal_set_applied_position'",
+							 true,
+							 0);
+		if (result != SPI_OK_SELECT)
+		{
+			ereport(FATAL,
+					(errmsg(TAG ": failed to detect "
+							"pgroonga_wal_set_applied_position(): "
+							"%u/%u: %d",
+							databaseOid,
+							tableSpaceOid,
+							result)));
+		}
+
+		if (SPI_processed > 0)
+		{
+			SetCurrentStatementStartTimestamp();
+			result = SPI_execute("SELECT pgroonga_wal_set_applied_position()",
+								 false,
+								 0);
+			if (result != SPI_OK_SELECT)
+			{
+				ereport(FATAL,
+						(errmsg(TAG ": failed to reset WAL applied positions "
+								"of all PGroonga indexes: "
+								"%u/%u: %d",
+								databaseOid,
+								tableSpaceOid,
+								result)));
+			}
+		}
+	}
+
+	PopActiveSnapshot();
+	SPI_finish();
+	CommitTransactionCommand();
+
+	pgstat_report_activity(STATE_IDLE, NULL);
+
+	proc_exit(0);
+}
+
 void
 pgroonga_crash_safer_reindex_one(Datum databaseInfoDatum)
 {
@@ -147,11 +219,11 @@ pgroonga_crash_safer_reindex_one(Datum databaseInfoDatum)
 	PushActiveSnapshot(GetTransactionSnapshot());
 	pgstat_report_activity(STATE_RUNNING, TAG ": reindexing");
 
-	pgrn_crash_safer_statuses_set_reindex_pid(NULL,
+	pgrn_crash_safer_statuses_set_prepare_pid(NULL,
 											  databaseOid,
 											  tableSpaceOid,
 											  MyProcPid);
-	before_shmem_exit(pgroonga_crash_safer_reindex_one_on_exit,
+	before_shmem_exit(pgroonga_crash_safer_prepare_one_on_exit,
 					  databaseInfoDatum);
 
 	{
@@ -290,6 +362,8 @@ pgroonga_crash_safer_flush_one(Datum databaseInfoDatum)
 	char pgrnDatabasePath[MAXPGPATH];
 	bool pgrnDatabasePathExist;
 	bool needReindex = false;
+	/* Only on the primary. */
+	bool needResetPosition = !RecoveryInProgress();
 	grn_ctx ctx;
 	grn_obj *db;
 	HTAB *statuses;
@@ -323,6 +397,7 @@ pgroonga_crash_safer_flush_one(Datum databaseInfoDatum)
 	{
 		grn_default_logger_set_path(PGroongaCrashSaferLogPath);
 	}
+	grn_set_default_n_workers(PGroongaCrashSaferMaxRecoveryThreads);
 
 	if (grn_init() != GRN_SUCCESS)
 	{
@@ -332,6 +407,7 @@ pgroonga_crash_safer_flush_one(Datum databaseInfoDatum)
 	}
 
 	grn_set_segv_handler();
+	grn_set_abrt_handler();
 
 	if (grn_ctx_init(&ctx, 0) != GRN_SUCCESS)
 	{
@@ -342,8 +418,14 @@ pgroonga_crash_safer_flush_one(Datum databaseInfoDatum)
 
 	GRN_LOG(&ctx,
 			GRN_LOG_NOTICE,
-			TAG ": initialize: <%s>",
-			PGRN_VERSION);
+			TAG ": initialize: <%s>: %u/%u",
+			PGRN_VERSION,
+			databaseOid,
+			tableSpaceOid);
+	GRN_LOG(&ctx,
+			GRN_LOG_DEBUG,
+			TAG ": max_recovery_threads: %d",
+			grn_get_default_n_workers());
 
 	grn_ctx_set_wal_role(&ctx, GRN_WAL_ROLE_PRIMARY);
 
@@ -376,14 +458,22 @@ pgroonga_crash_safer_flush_one(Datum databaseInfoDatum)
 	}
 	pfree(databasePath);
 
-	if (needReindex)
+	if (needReindex || needResetPosition)
 	{
 		BackgroundWorker worker = {0};
 		BackgroundWorkerHandle *handle;
 
+		GRN_LOG(&ctx,
+				GRN_LOG_NOTICE,
+				TAG ": %s: %u/%u",
+				needReindex ? "reindexing" : "resetting-position",
+				databaseOid,
+				tableSpaceOid);
+
 		snprintf(worker.bgw_name,
 				 BGW_MAXLEN,
-				 TAG ": reindex: %u/%u",
+				 TAG ": prepare: %s: %u/%u",
+				 needReindex ? "reindex" : "reset-position",
 				 databaseOid,
 				 tableSpaceOid);
 		snprintf(worker.bgw_type, BGW_MAXLEN, "%s", worker.bgw_name);
@@ -395,17 +485,30 @@ pgroonga_crash_safer_flush_one(Datum databaseInfoDatum)
 		snprintf(worker.bgw_library_name,
 				 BGW_MAXLEN,
 				 "%s", PGroongaCrashSaferLibraryName);
-		snprintf(worker.bgw_function_name,
-				 BGW_MAXLEN,
-				 "pgroonga_crash_safer_reindex_one");
+		snprintf(worker.bgw_function_name, BGW_MAXLEN,
+				 needReindex ?
+				 "pgroonga_crash_safer_reindex_one" :
+				 "pgroonga_crash_safer_reset_position_one");
 		worker.bgw_main_arg = databaseInfoDatum;
 		worker.bgw_notify_pid = MyProcPid;
 		if (RegisterDynamicBackgroundWorker(&worker, &handle))
 		{
 			WaitForBackgroundWorkerShutdown(handle);
+			GRN_LOG(&ctx,
+					GRN_LOG_NOTICE,
+					TAG ": %s: %u/%u",
+					needReindex ? "reindexed" : "reset-position",
+					databaseOid,
+					tableSpaceOid);
 		}
 	}
 
+	GRN_LOG(&ctx,
+			GRN_LOG_NOTICE,
+			TAG ": ready: %u/%u",
+			databaseOid,
+			tableSpaceOid);
+
 	statuses = pgrn_crash_safer_statuses_get();
 	pgrn_crash_safer_statuses_start(statuses, databaseOid, tableSpaceOid);
 	before_shmem_exit(pgroonga_crash_safer_flush_one_on_exit,
@@ -463,16 +566,77 @@ pgroonga_crash_safer_flush_one(Datum databaseInfoDatum)
 
 		/* TODO: How to implement safe finish on no connection? */
 		/*
-		if (pgrn_crash_safer_statuses_get_n_using_processing(statuses,
-															 databaseOid,
-															 tableSpaceOid) == 0)
+		if (pgrn_crash_safer_statuses_get_n_using_processes(statuses,
+															databaseOid,
+															tableSpaceOid) == 0)
 			break;
 		*/
 
 		grn_obj_flush_recursive(&ctx, db);
 	}
 
+	PGroongaCrashSaferGotSIGTERM = false;
+	while (!PGroongaCrashSaferGotSIGTERM && pgrn_file_exist(pgrnDatabasePath))
+	{
+		int conditions;
+		uint32 n_using_processes =
+			pgrn_crash_safer_statuses_get_n_using_processes(statuses,
+															databaseOid,
+															tableSpaceOid);
+		if (n_using_processes == 0)
+			break;
+
+		GRN_LOG(&ctx,
+				GRN_LOG_NOTICE,
+				TAG ": waiting for connections to finish: %u: %u/%u",
+				n_using_processes,
+				databaseOid,
+				tableSpaceOid);
+		conditions = WaitLatch(MyLatch,
+							   WL_LATCH_SET |
+							   WL_TIMEOUT |
+							   PGRN_WL_EXIT_ON_PM_DEATH,
+							   PGroongaCrashSaferFlushNaptime * 1000,
+							   PG_WAIT_EXTENSION);
+		if (conditions & WL_LATCH_SET)
+		{
+			ResetLatch(MyLatch);
+			CHECK_FOR_INTERRUPTS();
+		}
+
+		if (PGroongaCrashSaferGotSIGHUP)
+		{
+			PGroongaCrashSaferGotSIGHUP = false;
+			ProcessConfigFile(PGC_SIGHUP);
+		}
+	}
+
+	if (pgrn_file_exist(pgrnDatabasePath))
+	{
+		GRN_LOG(&ctx,
+				GRN_LOG_NOTICE,
+				TAG ": flushing database before closing: %u/%u",
+				databaseOid,
+				tableSpaceOid);
+		grn_obj_flush_recursive(&ctx, db);
+		GRN_LOG(&ctx,
+				GRN_LOG_NOTICE,
+				TAG ": flushed database before closing: %u/%u",
+				databaseOid,
+				tableSpaceOid);
+	}
+
+	GRN_LOG(&ctx,
+			GRN_LOG_NOTICE,
+			TAG ": closing database: %u/%u",
+			databaseOid,
+			tableSpaceOid);
 	grn_obj_close(&ctx, db);
+	GRN_LOG(&ctx,
+			GRN_LOG_NOTICE,
+			TAG ": closed database: %u/%u",
+			databaseOid,
+			tableSpaceOid);
 
 	grn_ctx_fin(&ctx);
 
@@ -527,7 +691,7 @@ pgroonga_crash_safer_main_flush_all(void)
 	HTAB *statuses;
 	const LOCKMODE lock = AccessShareLock;
 	Relation pg_database;
-	PGrnTableScanDesc scan;
+	TableScanDesc scan;
 	HeapTuple tuple;
 
 	StartTransactionCommand();
@@ -537,8 +701,8 @@ pgroonga_crash_safer_main_flush_all(void)
 
 	statuses = pgrn_crash_safer_statuses_get();
 
-	pg_database = pgrn_table_open(DatabaseRelationId, lock);
-	scan = pgrn_table_beginscan_catalog(pg_database, 0, NULL);
+	pg_database = table_open(DatabaseRelationId, lock);
+	scan = table_beginscan_catalog(pg_database, 0, NULL);
 	for (tuple = heap_getnext(scan, ForwardScanDirection);
 		 HeapTupleIsValid(tuple);
 		 tuple = heap_getnext(scan, ForwardScanDirection))
@@ -550,11 +714,7 @@ pgroonga_crash_safer_main_flush_all(void)
 		char pgrnDatabasePath[MAXPGPATH];
 		pgrn_crash_safer_statuses_entry *entry;
 
-#ifdef PGRN_FORM_PG_DATABASE_HAVE_OID
 		databaseOid = form->oid;
-#else
-		databaseOid = HeapTupleGetOid(tuple);
-#endif
 		tableSpaceOid = form->dattablespace;
 
 		databasePath = GetDatabasePath(databaseOid, tableSpaceOid);
@@ -571,8 +731,8 @@ pgroonga_crash_safer_main_flush_all(void)
 												 NULL);
 		pgroonga_crash_safer_main_flush_one(entry);
 	}
-	pgrn_table_endscan(scan);
-	pgrn_table_close(pg_database, lock);
+	table_endscan(scan);
+	table_close(pg_database, lock);
 
 	PopActiveSnapshot();
 	CommitTransactionCommand();
@@ -693,6 +853,21 @@ _PG_init(void)
 							 NULL,
 							 NULL);
 
+	DefineCustomIntVariable("pgroonga_crash_safer.max_recovery_threads",
+							"Maximum number of threads for recovery of broken Groonga indexes.",
+							"The default is 0, which means disabled. "
+							"Use all CPUs in the environment at -1. "
+							"Use CPU for that number if 1 or later is set.",
+							&PGroongaCrashSaferMaxRecoveryThreads,
+							PGroongaCrashSaferMaxRecoveryThreads,
+							-1,
+							INT_MAX,
+							PGC_SIGHUP,
+							0,
+							NULL,
+							NULL,
+							NULL);
+
 	if (!process_shared_preload_libraries_in_progress)
 		return;
 
